/*
  rdesktop: A Remote Desktop Protocol client.
  User interface services - Amiga
  Copyright (C) Michal Zukowski 2012-2018
  Copyright (C) Martin Blom 2001-2005
  Copyright (C) Matthew Chapman 1999-2000

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <string.h>
#define USE_INLINE_STDARG 1
#include <proto/charsets.h>
#include "rdesktop.h"

/* Output a string in Unicode */
void rdp_out_unistr(STREAM s, char *string, int len)
{ 
	/* LEGACY:
	 *
	 *  Do not write new code that uses this function, use the ones defined
	 *  in stream.h for writing utf16 strings to a stream.
	 *
	 */
	
	size_t ibl, obl;
	char *pin, *pout;


	if (string == NULL || len == 0)
		return;
	
	ibl = strlen(string);
	obl = len + 2;
	pin = string;
	pout = (char *) s->p;

	memset(pout, 0, len + 4);

	
	ConvertTags((STRPTR)pin, -1, pout, obl, MIBENUM_SYSTEM, MIBENUM_UTF_16LE, TAG_DONE);

	s->p += len + 2;
}


/* Input a string in Unicode
 *
 * Returns str_len of string
 */
void rdp_in_unistr(STREAM s, int in_len, char **string, uint32 * str_size)
{
	/* Corner case. We still want to return a null terminated string... */
	if (in_len == 0) 
	{	
		if (*string == NULL)
		{
			*string = xmalloc(1);
		}
		**string = '\0';
		*str_size = 0;
		return;
	}
	else
	{
		/* Dynamic allocate of destination string if not provided */
		if (*string == NULL)
		{
			*string = xmalloc(in_len/2);
			*str_size = in_len;
			//KPrintF("allocated\n");
		}
		
		*str_size = ConvertTags(s->p, in_len, *string,  *str_size, MIBENUM_UTF_16LE, MIBENUM_SYSTEM, TAG_DONE);
		
		// we must update the location of the current STREAM for future reads of s->p 
		s->p += in_len;
	}
}


/* Writes a utf16 encoded string into stream excluding null termination.
   This function assumes that input is ASCII compatible, such as UTF-8.
 */
size_t _out_utf16s(STREAM s, size_t maxlength, const char *string)
{
  size_t ibl, obl;
  const char *pin;
  char *pout;

  if (string == NULL)
    return 0;
  
	 ibl = strlen(string);
  obl = maxlength ? maxlength : (size_t)s_left(s);
  pin = string;
  pout = (char *) s->p;

	ConvertTags((STRPTR)string, ibl, pout, obl, MIBENUM_SYSTEM, 
		MIBENUM_UTF_16LE,  CST_DoNotTerminate, TRUE, TAG_DONE);

  s->p = (unsigned char *)pout+2*ibl;
    return 2*ibl;
}

char *ConvertToSystemAlloc(uint8 * data, uint32 length, uint32 *dst_length)
{
	char *dst_str = NULL;
	
	if (ConvertTags((STRPTR)data, length, NULL, -1, MIBENUM_UTF_16LE, MIBENUM_SYSTEM,  
		CST_GetDestBytes, (IPTR)dst_length, TAG_DONE) != -1)
	{
		if ((dst_str = AllocVec(*dst_length, MEMF_ANY)))
		{
			ConvertTags((STRPTR)data, length, dst_str, *dst_length, MIBENUM_UTF_16LE, MIBENUM_SYSTEM, TAG_DONE);
		}
	}
	return dst_str;
}

char *ConvertToUTF8Alloc(uint8 * data, uint32 length, uint32 *dst_length)
{
	char *dst_str = NULL;
	
	if (ConvertTags((STRPTR)data, length, NULL, -1, MIBENUM_UTF_16LE, MIBENUM_SYSTEM,  
		CST_GetDestBytes, (IPTR)dst_length, TAG_DONE) != -1)
	{
		if ((dst_str = AllocVec(*dst_length, MEMF_ANY)))
		{
			ConvertTags((STRPTR)data, length, dst_str, *dst_length, MIBENUM_UTF_16LE, MIBENUM_SYSTEM, TAG_DONE);
		}
	}
	return dst_str;
}



char *ConvertToUTF16Alloc(uint8 * data, uint32 length, uint32 *dst_length)
{
	char *dst_str = NULL;
	
	if (ConvertTags((STRPTR)data, length, NULL, -1, MIBENUM_SYSTEM, MIBENUM_UTF_16LE,
		CST_GetDestBytes, (IPTR)dst_length, TAG_DONE) != -1)
	{
		if ((dst_str = AllocVec(*dst_length, MEMF_ANY)))
		{
			ConvertTags((STRPTR)data, length, dst_str, *dst_length, MIBENUM_SYSTEM, MIBENUM_UTF_16LE, TAG_DONE);
		}
	}
	return dst_str;
}



char *ConvertUTF8ToUTF16Alloc(uint8 * data, uint32 length, uint32 *dst_length)
{
	char *dst_str = NULL;
	
	if (ConvertTags((STRPTR)data, length, NULL, -1, MIBENUM_UTF_8, MIBENUM_UTF_16LE,
		CST_GetDestBytes, (IPTR)dst_length, TAG_DONE) != -1)
	{
		if ((dst_str = AllocVec(*dst_length, MEMF_ANY)))
		{
			ConvertTags((STRPTR)data, length, dst_str, *dst_length, MIBENUM_UTF_8, MIBENUM_UTF_16LE, TAG_DONE);
		}
	}
	return dst_str;
}
